"""
Computer Control MCP - Python package for computer control via MCP.

This package provides computer control capabilities using PyAutoGUI through a
Model Context Protocol (MCP) server.
"""

from computer_control_mcp.core import mcp, main

__version__ = "0.1.2"
__all__ = ["mcp", "main"]
